'' Common definitions and persistent storage allocation

'' 2007-02-26 - Initial revisions for use with BOE-BOT Basic
'' 2007-09-29 - Modified for configurable serial buffer sizes
'' 2007-11-25 - Added definitions for boot serial I/O pins
'' 2007-12-20 - Added definitions for keyboard and display

CON
'' Hardware dependent and other basic definitions

  clockMode   = xtal1 | pll16x
  xtalClock   = 5_000_000
  spiDO       =  8
  spiClk      =  9                             ' I/O pins for SPI interface to SD card
  spiDI       = 10
  spiCS       = 11
  vgaPins     = 16                             ' I/O pins for VGA display (8 pins)
  keyPins     = 26                             ' I/O pins for PS/2 keyboard (2 pins)
  bootTx      = 30                             ' Serial downloading I/O pins for debugging
  bootRx      = 31
 
'' General hardware dependent definitions

  maxPinPairs = 16                             ' Maximum number of possible I2C busses
  i2cBootSCL  = 28                             ' I2C Clock pin for boot EEPROM
  i2cBootSDA  = i2cBootSCL + 1                 ' I2C Data pin for boot EEPROM
  bootAddr    = i2cBootSCL << 18               ' File address for boot EEPROM

'' The following are fixed locations assigned at the top of RAM memory to avoid
'' having them overwritten when a new program is loaded.  Most of these are pointers
'' to work areas allocated "below" them.  The last pointer is reserved for non-system
'' programs

  endMemory   = $8000                          ' Address past the end of RAM
  initMarker  = endMemory -   4                ' Unique marker for initialized data
  memPtr      = endMemory -   8                ' Address of last allocated long
  randomSeed  = endMemory -  12                ' Random seed for "?" operator
  ioControl   = endMemory -  20                ' I/O control block for general use
  loaderCog   = endMemory -  24                ' Cog number + 1 of loader cog
  keyBreak    = endMemory -  28                ' Break key code, flag  read/write
  keyKeys     = endMemory -  60                ' Key buffer (16 words) read-only
  keyStates   = endMemory -  92                ' Key states (256 bits) read-only
  keyPresent  = endMemory -  96                ' Keyboard present      read-only
  keyHead     = endMemory - 100                ' Key buffer head       read-only
  keyTail     = endMemory - 104                ' Key buffer tail       read/write
  keyCog      = endMemory - 108                ' Cog number + 1 of keyboard cog
  dspMode     = endMemory - 112                ' Mode: 0 - 16x16 tiles, 1 - 16x32 tiles
  dspSync     = endMemory - 116                ' Display sync flag
  dspCursor   = endMemory - 120                ' Pointer to cursor information
  dspColor    = endMemory - 124                ' Pointer to color table
  dspArray    = endMemory - 128                ' Pointer to tile pointers
  dspCnt      = endMemory - 132
  dspVcfg     = endMemory - 136
  dspDirb     = endMemory - 140
  dspDira     = endMemory - 144
  dspCog      = endMemory - 148                ' Cog number + 1 of display cog(s)
  dspRows     = endMemory - 152                ' Number of rows for display
  dspCols     = endMemory - 156                ' Number of columns for display
  dspTiles    = endMemory - 160                ' Number of tiles in display
  systemPtr   = endMemory - 164                ' Address of last allocated system area
  userPtr     = endMemory - 168                ' Address of work area for "user"
  endFree     = userPtr                        ' Address past initial free memory

  uniqueMark  = $965A3EC1                      ' Indicates that RAM was initialized
  noSuchAddr  = $8000                          ' "Invalid" address for initializing pointers

'' Command definitions for i2cSpiLdr routines

  ioReadCmd   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  ioWriteCmd  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  ioRead1Cmd  = %00000011                      ' Read from a device with only 8-bit addresses
  ioWrite1Cmd = %00000100                      ' Write to a device with only 8-bit addresses
  ioRead0Cmd  = %00000101                      ' Read from a device without address bytes
  ioWrite0Cmd = %00000110                      ' Write to a device without address bytes
  ioBootCmd   = %00001000                      ' Read from EEPROM to HUB RAM, then start a
                                               '  new SPIN interpreter in the COG whose ID is
                                               '  supplied in the lower 3 bits of this command
                                               '  This COG is stopped before the read is done.
  ioSpiInit   = %00010000                      ' Initialize the specified SPI bus and SD card
  ioSpiStop   = %00010001                      ' Change all SD card pins to inputs
  ioSpiRead   = %00010010                      ' Read one or more bytes from the SD card
  ioSpiWrite  = %00010011                      ' Write one or more bytes from the SD card
  ioSpiBoot   = %00011000                      ' Like ioBootCmd, but uses ioSpiRead for loading
  ioCmdMask   = %00011111                      ' Used to mask off command bits
  ioSpiMask   = %00010000                      ' Used to test for SPI command codes

' Options for commands

  ioNoStore   = %00100000                      ' If set, data is not stored into main memory
                                               '  If ioBootCmd or ioSpiBoot, no cogs are
                                               '   stopped and a new cog is not started.
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioStopLdr   = %10000000                      ' If set, the loader's cog is stopped after a boot

' Return status

  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

'' Internal SPIN information at fixed locations in RAM/EEPROM image

  clkfreqVal  = $0000                          ' Current CLKFREQ value stored here
  clksetVal   = $0004                          ' Current CLKSET value stored here
  chksumVal   = $0005                          ' Checksum over memory stored here
  spinPbase   = $0006                          ' must be $0010
  spinVbase   = $0008                          ' number of longs loaded times 4
  spinDbase   = $000A                          ' above where $FFF9FFFF's get placed
  spinPcurr   = $000C                          ' points to SPIN code
  spinDcurr   = $000E                          ' points to local stack
  
'' Display control codes

  ClrScr      = $00                            ' clear screen  
  Home        = $01                            ' home in screen
  MoveXY      = $02                            ' move to X,Y in screen (X and Y follow)
  CursLt      = $03                            ' cursor left
  CursRt      = $04                            ' cursor right
  CursUp      = $05                            ' cursor up
  CursDn      = $06                            ' cursor down
  Color       = $07                            ' select color C (0-63) (C follows)
  Bsp         = $08                            ' backspace
  Tab         = $09                            ' tab (8 spaces per)
  Lf          = $0A                            ' linefeed
  ClrEol      = $0B                            ' clear to end of line
  Insert      = $0C                            ' insert blank at X,Y
  Cr          = $0D                            ' carriage return
  MoveX       = $0E                            ' move to X in window (X follows)
  MoveY       = $0F                            ' move to Y in window (Y follows)
  Esc         = $1B                            ' (ESC) print next character C as-is (C follows)
  Quote       = $22                            ' ASCII double quote
  Delete      = $7F                            ' delete current character
  Alt00       = $FF                            ' changed to $00 by out (for use in string())

'' Parameters for Keyboard to set buffer size

CON
  keySize  = 1 << 4              ' Size of keyboard buffer
  keyMask  = keySize - 1         ' Mask for keyboard buffer size

'' Parameters for FullDuplexSerial to set buffer size
'' Currently: 64 bytes for receive and 64 bytes for transmit

  rxBufSize   = 1 << 6                         ' Receive buffer size (must be power of 2)
  rxPtrMask   = rxBufSize - 1                  ' Receive buffer pointer mask
  txBufSize   = 1 << 6                         ' Transmit buffer size (must be power of 2)
  txPtrMask   = txBufSize - 1                  ' Transmit buffer pointer mask

PUB start                                      ' Initialize the high memory area if not done
   ifnot resumed
      longfill(endFree,0,(endMemory-endFree)/4)
      long[systemPtr] := endFree               ' Next available system work area address
      deallocate
      long[initMarker] := uniqueMark           ' Marks high memory area as initialized
      long[randomSeed] := uniqueMark           ' Initial random seed value

PUB resumed                                    '' Return true if already initialized
   return long[initMarker] == uniqueMark

PUB allocatePRI(size)                          '' Allocate block of size bytes
   return long[systemPtr] := long[memPtr] := (long[memPtr] - size) & !3

PUB allocate(size)                             '' Allocate a block of user work area
   if long[userPtr] <> noSuchAddr
     deallocate                                ' Deallocate current area if allocated
   return long[userPtr] := long[memPtr] := (long[memPtr] - size) & !3

PUB deallocate                                 '' User work area no longer in use
   long[memPtr] := systemPtr
   long[userPtr] := noSuchAddr                 ' Use ROM address for "invalid" address