'' SD Card support based heavily on Rokicki's FAT file system
'' Modified by Michael Green to support multiple instances and
'' to allow files on EEPROM 32K pages as well.  "pass through"
'' calls to low level SPI and I2C I/O calls also provided.

'' 2008-01-01 - Initial implementation
'' 2008-01-06 - Added support for file seek and update in place.
''              A file can be truncated at the current file position.
''              EEPROM files now have a file length stored with the
''              file name so they can use append mode.

'' The low level I2C / SPI driver (I2cSpiInit) has to be initialized before
'' initializing this object by calling mountSDVol.  Call its start routine.
'' The mountSDVol routine is called whenever a new SD card is to be mounted.
'' unmountSDVol can be called when an SD card is unmounted, but this is not
'' necessary.  All files must be closed though before an SD card is removed.

'' The mountSDVol, and unmountSDVol routines update all instances of this
'' object and should not be called when a file is open.  Otherwise, one
'' file per instance can be opened.

'' Note that EEPROM-based files can be used.  These are filenames that start
'' with a colon (:).  They reside in 32K EEPROM blocks in EEPROM attached to
'' the boot EEPROM pins of the Propeller (I/O pins 28 & 29).  The normal boot
'' area has the fixed name ":BOOT" which cannot be deleted.  The file name
'' file size is stored in the last 16 bytes of each 32K EEPROM block except
'' the boot area where the name is assumed and the file size is always 32K-16.

OBJ
   def  : "Definitions"         ' Propeller OS definitions
   ldr  : "i2cSpiLdr"           ' Loader / I2C & SPI routines

CON
   eepromPage    = 64           ' EEPROM page size
   
   bootName1     = $4F4F423A    ' Boot EEPROM name ":BOO", "T   "
   bootName2     = $20202054
   
   blankName     = $20202020    ' Blanks
   periodName    = $2020202E    ' A single period
   goUpName      = $20202E2E    ' Two periods
   nameAddr      = $7FF0        ' EEPROM address of name in 32K page
   sizeAddr      = $7FFC        ' File size, then ?fileSize check value

   sectorShift   = 9
   sectorSize    = |< sectorShift
   sectorMask    = sectorSize - 1
   dirShift      = 5
   dirSize       = |< dirShift
   eepromSize    = |< 15        ' 32K
   eepromMask    = eepromSize - 1

DAT
readSDErr        byte  "Can't read from SD card",0
writeSDErr       byte  "Can't write to SD card",0
mountSDErr       byte  "Can't mount SD card",0
fat16Err         byte  "Not a FAT16 volume",0
sectorErr        byte  "Sector size not 512",0
clusterErr       byte  "Cluster size < 2 or not 2^n",0
fatCountErr      byte  "Must have two FATs",0
fatMarkErr       byte  "Bad FAT signature",0
tooBigErr        byte  "Too many clusters",0
badClusterErr    byte  "Invalid cluster number",0
notWriteErr      byte  "Write in read mode",0
noSpaceErr       byte  "No space left on volume",0
badAddrErr       byte  "Invalid seek address",0
protectedErr     byte  "File write protected",0
openModeErr      byte  "Open mode not r/w/a/d",0
dirFullErr       byte  "Directory full",0
bootModeErr      byte  ":BOOT cannot be deleted",0
noEepromErr      byte  "No EEPROM present",0
noAppendErr      byte  "No EEPROM append mode",0
writeEepromErr   byte  "Write EEPROM error",0
writeWaitErr     byte  "Wait EEPROM error",0
noEmptyErr       byte  "EEPROM directory full",0
readEepromErr    byte  "Read EEPROM error",0
fileSizeErr      byte  "Invalid EEPROM file size",0
subDirSizeErr    byte  "Subdirectory size invalid",0

fat16Mark        long  "F"+("A"<<8)+("T"<<16)+("1"<<24)

rootDir          long  0         ' Byte address of start of root dir
rootDirEnd       long  0         ' Byte address following root dir
dataRegion       long  0         ' Block address of data region
clusterShift     long  0         ' Log base 2 of blocks per cluster
clusterSize      long  0         ' Number of blocks per cluster
clusterMask      long  0         ' Mask for cluster number
clusterByteShift long  0         ' Log base 2 of bytes per cluster
clusterByteSize  long  0         ' Number of bytes per cluster
clusterByteMask  long  0         ' Mask for bytes per cluster
fat1             long  0         ' Block address of the first FAT
totalClusters    long  0         ' Total clusters on the volume
sectorsPerFAT    long  0         ' Total sectors per FAT
dirListPtr       long  0         ' Directory byte address for nextSDFiles
dirListLen       long  0         ' Directory length to go for nextSDFiles

lastRead         long  0         ' Block address of block in metaBuf
dirty            long  0         ' True if metaBuf has been changed

metaBuf          long  0 [sectorSize/4] ' Buffer used for metaData

VAR
   long fclust                   ' Current cluster number
   long fileSize                 ' Total byte size of file
   long floc                     ' Current sector address in file
   long frem                     ' Current remaining bytes in cluster
   long bufat                    ' Current position within buffer
   long bufend                   ' End of data in buffer
   long writeLink                ' Byte address to store new cluster
   long dirEntry                 ' Byte address of directory entry
   long fatPtr                   ' Byte address of most recent FAT entry
   long eepromPtr                ' -1 or EEPROM address for EEPROM files
   long backupChar               ' Single character backup for getLine
   long readMode                 ' File opened in "r" mode.  No writing.
   long subDirectory             ' Sector address of current subdirectory
   
   long buffer[sectorSize/4]     ' Data buffer

PUB mountSDVol | s, t            ' Mount SD card & initialize driver
   if ldr.initSDCard(def#SpiDO,def#SpiClk,def#SpiDI,def#SpiCS)
      abort -@mountSDErr
   eepromPtr := -1               ' Default I/O is SPI
   readMode := true
   backupChar := -1
   lastRead~~                    ' No sector currently in buffer
   dirty~
   readMeta(0)                   ' Read boot block to meta buffer
   if metaReadLong($36) == fat16Mark
      s := 0
   else
      readMeta(s := metaReadLong($1c6))
   if metaReadLong($36) <> fat16Mark or metaBuf.byte[$3a] <> "6"
      abort -@fat16Err           ' Not a FAT16 volume
   if metaReadWord($0b) <> SectorSize
      abort -@sectorErr          ' Sector size not 512
   clusterSize := metaBuf.byte[$0d]
   clusterShift := lookdown(clusterSize:2,4,8,16,32,64,128)
   if clusterSize < 2 or clusterShift == 0
      abort -@clusterErr         ' Cluster size < 2 or not 2^n
   clusterMask := clusterSize - 1
   clusterByteShift := sectorShift + clusterShift
   clusterByteSize := |< clusterByteShift
   clusterByteMask := clusterByteShift - 1
   if metaBuf.byte[$10] <> 2
      abort -@fatCountErr        ' Must have 2 FATs
   ifnot t := metaReadWord($13)
      t := metaReadLong($20)     ' Get total sectors on volume
   sectorsPerFAT := metaReadWord($16)
   if metaReadWord($1fe) <> $aa55
      abort -@fatMarkErr         ' Bad FAT signature
   fat1 := s + metaReadWord($0e) ' Start + Reserved
   rootDir := (fat1 + 2*sectorsPerFAT) << SectorShift
   rootDirEnd := rootDir + (metaReadWord($11) << dirShift)
   dataRegion := fat1 + 2*sectorsPerFat + (rootDirEnd - rootDir) >> sectorShift
   dataRegion -= 2 * clusterSize ' Adjust for clusters starting at 2
   totalClusters := (t + s - dataRegion) >> clusterShift
   if totalClusters > $fff0
      abort -@tooBigErr          ' Too many clusters
   subDirectory := 0             ' Root directory

PUB debugAddr
   return @rootDir << 16 | @fclust ' Return addresses for debugging
   
PUB unmountSDVol                 ' Unmount SD card & disconnect driver
   return ldr.stopSDCard

PUB startSDFiles                 ' Start current directory listing
   if subDirectory
      dirListPtr := (subDirectory + dataRegion) << sectorShift
      dirListLen := clusterByteSize
   else
      dirListPtr := rootDir
      dirListLen := rootDirEnd - rootDir

PUB nextSDFiles(fbuf) | i, p     ' Get next file name (put in fbuf)
   repeat while dirListLen > 0
      readMeta(dirListPtr >> sectorShift)
      p := dirListPtr & sectorMask
      if metaBuf.byte[p] == 0    ' End of directory?
         return false
      if metaBuf.byte[p] <> $e5 and metaBuf.byte[p+$0b]&$18 == 0
         repeat i from 0 to 10   ' If valid directory entry, copy name
            if (byte[fbuf] := metaBuf.byte[p+i]) <> " "
               fbuf++            ' Skip blanks and insert
            if i == 7            '  period when needed
               byte[fbuf++] := "."
         byte[fbuf]~             ' Zero terminated string
         return true
      dirListPtr += dirSize
      dirListLen -= dirSize
   return false

PUB openFile(s,mode) | p,dt,n1,n2,n3,t ' Open file s.  Mode r/w/a/d
   closeFile
   n1 := n2 := n3 := blankName   ' Scan 8.3 file name
   p := 0
   repeat while t := byte[s++]   ' Get up to 8 characters delimited
      if t == "."                '  by either a period or zero byte
         quit
      if p < 8
         n1.byte[p++] := uc(t)
   if t == "."                   ' If a period found, get up to 3
      p := 0                     '  characters delimited by a zero
      repeat while t := byte[s++]
         if p < 3
            n3.byte[p++] := uc(t)
   if n1 == bootName1 and n2 == bootName2 and n3 == blankName
      if mode == "d"             ' :BOOT cannot be deleted
         abort -@bootModeErr
      if ldr.checkPresence(def#bootAddr)
         eepromPtr := def#bootAddr
         fileSize := nameAddr
         return                  ' Make sure boot EEPROM present
      abort -@noEepromErr
   if n1.byte == ":"             ' Leading ":" is a 32K EEPROM file
      if (p := lookupName(n1,n2,n3)) <> -1
         if mode == "d"          ' Delete mode.  Zero file name area
            n1 := n2 := n3 := t := 0
            if ldr.writeEEPROM(p+nameAddr,@n1,16)
               abort -@writeEepromErr
            if ldr.writeWait(p+nameAddr)
               abort -@writeWaitErr
         else
            readMode := mode == "r"
            eepromPtr := p       ' Read, Write, or Append mode
            if ldr.readEEPROM(p+sizeAddr,@t,4)
               abort -@readEepromErr
            fileSize := t.word[0]
            if t.word[1] <> ?t.word[0] ' Make sure file size is valid
               abort -@fileSizeErr
            if mode == "a"       ' Move to end of file on append
               eepromPtr += fileSize
         return
      if mode == "w" or mode == "a" ' Write or Append mode with new file
         if (p := lookupName(0,0,0)) <> -1
            fileSize := t := 0   ' Set file size to zero
            ?t.word[1]           '  (include check value)
            if ldr.writeEEPROM(p+nameAddr,@n1,16)
               abort -@writeEepromErr ' Write filename to EEPROM area
            if ldr.writeWait(p+nameAddr)
               abort -@writeWaitErr
            readMode := false
            eepromPtr := p
            return
         abort -@noEmptyErr
      abort -@openModeErr
   if directoryLookup(@n1,false) ' SD card root directory lookup
      p := dirEntry & sectorMask
      readMode := mode == "r"
      if (metaBuf.byte[p+$0b] & $d9) <> 0 and not readMode
         abort -@protectedErr    ' File write protected
      case mode
         "r":                    ' Read file mode
            frem := sectorSize << clusterShift <# fileSize
         "w":                    ' Write new file mode
            metaWriteWord(p+$1a,-1) ' First cluster number
            metaWriteLong(p+$1c,0) ' File size
            dt := dateTime
            metaWriteLong(p+$16,dt) ' Last modified date/time
            writeLink := p + $1a ' Save offset to first cluster
            freeClusters(fclust)
            flushMeta
            bufend := sectorSize ' Now open new empty file
            fclust := 0
            fileSize := 0
            frem := 0
         "a":                    ' Append to existing or new file
            seekFile(fileSize)
         "d":                    ' Delete file
            metaBuf.byte[p] := $e5
            dirty~~              ' Mark specified file as
            freeClusters(fclust) '  deleted and free up space
            flushMeta
         other:
            abort -@openModeErr  ' Open mode not r/w/a/d
   else
      if mode == "r" or mode == "d"
         return -1               ' File not found
      if dirEntry == 0
         abort -@dirFullErr      ' Directory full
      p := dirEntry & sectorMask
      readMode := false
      readMeta(dirEntry >> sectorShift)
      t := metaBuf.byte[p] == 0  ' Are we adding new directory entry?
      bytefill(@metaBuf+p,0,dirSize)
      bytemove(@metaBuf+p,@n1,11) ' Copy file name
      metaWriteWord(p+$1a,-1)    ' First cluster number
      dt := dateTime
      metaWriteLong(p+$0e,dt)    ' Set create time and date
      metaWriteLong(p+$16,dt)    ' Set last modified date and time
      if t and dirEntry + dirSize < rootDirEnd ' Mark end of directory
         readMeta((dirEntry + dirSize) >> sectorShift)
         metaBuf.byte[(dirEntry + dirSize) & sectorMask]~
         dirty~~
      flushMeta
      writelink := dirEntry + $1a ' Save offset to first cluster
      fclust := 0
      bufend := sectorSize

PUB closeFile | t                ' Close currently opened file
   if eepromPtr => 0
      if not readMode            ' Update file size if not read mode
         t.word[0] := t.word[1] := fileSize
         ?t.word[1]              ' Save current file size to EEPROM
         eepromPtr &= !eepromMask
         if ldr.writeEEPROM(eepromPtr+sizeAddr,@t,4)
            abort -@writeEepromErr
         if ldr.writeWait(eepromPtr+sizeAddr)
            abort -@writeWaitErr
      eepromPtr := -1
   else
      if not readMode            ' If file open for writing,
         result := flushBuffer(true) ' flush last buffer if any
      bufat := bufend := 0
      floc := frem := fclust := 0
      writeLink := 0
      dirEntry := 0
   readMode := true
   fileSize := 0
   backupChar := -1

PUB seekFile(address) | t        ' Position the file as requested
   if address == -1
      address := fileSize        ' Use current file size if -1 given
   backupChar := -1
   if address < 0 or address > fileSize
      abort -@badAddrErr         ' Invalid seek address
   if eepromPtr <> -1
      eepromPtr := address       ' For EEPROM, just change position
   else
      frem := address
      readMeta(dirEntry >> sectorShift)
      fclust := metaReadWord(dirEntry & !sectorMask + $1a)
      if fclust < 2              ' Next cluster cannot be invalid
         abort -@badClusterErr
      if fclust => $fff0         ' If at end of file, try again from
         fclust := 0             '  beginning of file to find position
      repeat while frem > t 
         fclust := nextCluster   ' Find cluster for file position
         if fclust < 2           ' Invalid cluster following chain
            abort -@badClusterErr
         frem -= clusterByteSize
      floc := frem & !sectorMask ' Position to sector and offset
      bufat := frem & sectorMask '  within the file cluster
      frem := fileSize - address <# clusterByteSize - frem ' Space left in current cluster
      if readMode
         bufend := fileSize - floc <# sectorSize ' Adjust for file size
      else
         bufend := sectorSize    ' For writes, full buffer available
      if address < fileSize or address == fileSize and bufat <> 0
         ' If start of sector not at end of file, read in the sector
         if ldr.readSDCard(getDatablock,@buffer,sectorSize)
            abort -@readSDErr
      if fillDataBuffer == -1
      ' ???
   
PUB readFile(address,count) | t  ' Read a block from the current file
   backupChar := -1              '  and return the number of bytes read
   if eepromPtr <> -1
      if ldr.readEEPROM(eepromPtr,address,count)
         abort -@readEepromErr
      eepromPtr += count
      return count
   repeat while count > 0
      if bufat => bufend
         if fillDataBuffer =< 0
            if result == 0
               return -1
            else
               return
      t := count <# bufend - bufat
      bytemove(address,@buffer+bufat,t)
      bufat += t
      address += t
      count -= t
      result += t

PUB getByte                      ' Read a byte from the current file
   if eepromPtr <> -1
      if ldr.readEEPROM(eepromPtr,@result,1)
         abort -@readEepromErr
      eepromPtr++
   else
      if bufat => bufend         ' If buffer is empty,
         if fillDataBuffer =< 0  '  go get a new buffer
            return -1
      result := buffer.byte[bufat++]

PUB getLine(p,size) | c, onlyEof ' Read a line of text from the current file
   onlyEof := true               '  and return true if an EOF line is present
   repeat                        ' "p" is the address of a byte array to hold
      if backupChar <> -1        '  the line.  Maximum length of the line is
         result := backupChar~~  '   "size" excluding the zero byte terminator
      else
         result := getByte
      case result
         -1, 0:                  ' End of file
            byte[p] := 0
            backupChar~
            return onlyEof
         def#Cr:
            onlyEof := false
            case backupChar := getByte
               -1, 0:
                  backupChar~
               def#Lf:
                  backupChar~~
               other:
                  next
            byte[p] := 0
            return false
         def#Lf:
            byte[p] := 0
            return false
         other:
            if size
               byte[p++] := result
               size--
            onlyEOF := false

PUB writeFile(address,count) | t ' Write a block to the current file
   if eepromPtr <> -1
      result := count            ' Handle first partial page
      if count > (t := eepromPage - eepromPtr & (eepromPage-1))
         if ldr.writeEEPROM(eepromPtr,address,t)
            abort -@writeEepromErr
         if ldr.writeWait(eepromPtr)
            abort -@writeWaitErr
         eepromPtr += t
         address += t
         count -= t
      repeat while count > eepromPage ' Handle full pages
         if ldr.writeEEPROM(eepromPtr,address,count)
            abort -@writeEepromErr
         if ldr.writeWait(eepromPtr)
            abort -@writeWaitErr
         eepromPtr += eepromPage
         address += eepromPage
         count -= eepromPage
      if ldr.writeEEPROM(eepromPtr,address,count)
         abort -@writeEepromErr   ' Handle any remainder
      if ldr.writeWait(eepromPtr)
         abort -@writeWaitErr
      eepromPtr += count
   else
      repeat while count > 0
         if bufat => bufend
            flushBuffer(false)
         t := count <# bufend - bufat
         bytemove(@buffer+bufat,address,t)
         bufat += t
         address += t
         count -= t

PUB putByte(c)                   ' Write a byte to the current file
   if eepromPtr <> -1
      if ldr.writeEEPROM(eepromPtr,@c,1)
         abort -@writeEepromErr
      if ldr.writeWait(eepromPtr)
         abort -@writeWaitErr
      eepromPtr++
   else
      buffer.byte[bufat++] := c
      if bufat == sectorSize
         return flushBuffer(false)

PUB putLine(s)                   ' Write a text line to the current file
   writeFile(s,strsize(s))
   writeFile(string(def#Cr,def#Lf),2) ' Use Cr/Lf as a line delimiter

PUB bootFile(s)                  ' Boot from the specified file
   if (result := openFile(s,"r")) == 0
      if eepromPtr <> -1
         result := ldr.bootEEPROM(eepromPtr)
      else
         ' Make sure filesize =< cluster size
         result := ldr.bootSDCard(getDatablock, filesize <# $8000)
      closeFile

PUB goSubdirectory(s) | p,n1,n2,n3,t ' Go to specified subdirectory
   n1 := n2 := n3 := blankName   ' Scan 8.3 file name
   p := 0
   repeat while t := byte[s++]   ' Get up to 8 characters delimited
      if t == "."                '  by either a period or zero byte
         quit
      if p < 8
         n1.byte[p++] := uc(t)
   if t == "."                   ' If a period found, get up to 3
      p := 0                     '  characters delimited by a zero
      repeat while t := byte[s++]
         if p < 3
            n3.byte[p++] := uc(t)
   if n1 == blankName and n2 == blankName and n3 == blankName
      n1 := periodName           ' Special case of "."
   if n1 == blankName and n2 == blankName and n3 == periodName
      n1 := goUpName             ' Special case of ".."
      n3 := blankName
   if directoryLookup(@n1,true)  ' SD card subdirectory lookup
      p := dirEntry & sectorMask
      if fileSize <> 0           ' Invalid subdirectory
         abort -@subDirSizeErr
      t := metaReadWord(p + $1a) ' Return data sector # for cluster
      if t => 2 and t < totalclusters
         return t << clusterShift

PUB readExpander(address)        ' Return a byte from a PCF9554 I/O Expander
   return ldr.readExpander(address)

PUB writeExpander(address, value) ' Write a byte to a PCF9554 I/O Expander
   return ldr.writeExpander(address, value)

PUB readEEPROM(address, bufAdr, size) ' Low level read EEPROM
   return ldr.readEEPROM(address, bufAdr, size)

PUB writeEEPROM(address, bufAdr, size) ' Low level write EEPROM
   return ldr.writeEEPROM(address, bufAdr, size)

PUB checkPresence(address)       ' Low level check for presence of EEPROM
   return ldr.checkPresence(address)

PUB writeWait(address)           ' Low level wait for write EEPROM completion
   return ldr.writeWait(address)

PUB bootEEPROM(address)          ' Low level boot from EEPROM
   return ldr.bootEEPROM(address)

PUB readSDCard(address, bufAdr, size) ' Low level read SD card
   return ldr.readSDCard(address, bufAdr, size)

PUB writeSDCard(address, bufAdr, size) ' Low level write SD card
   return ldr.writeSDCard(address, bufAdr, size)

PUB bootSDCard(address, size)    ' Low level boot SD card
   return ldr.bootSDCard(address, size)
  
PUB out(c)                       ' Output a single character
   putByte(c)

PUB str(stringptr)               ' Output a zero-terminated string
   return writeFile(stringptr,strsize(stringptr))

PUB dec(value) | i               ' Output a decimal value
  if value < 0
    -value
    putByte("-")
  i := 1_000_000_000
  repeat 10
    if value => i
      putByte(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      putByte("0")
    i /= 10
    
PUB hex(value, digits)           ' Output a hexadecimal value
  value <<= (8 - digits) << 2
  repeat digits
    putByte(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)           ' Output a binary value
  value <<= 32 - digits
  repeat digits
    putByte((value <-= 1) & 1 + "0")

PUB getDatablock                 ' Return block address of data block
   return (fclust<<clusterShift)+dataRegion+((floc>>sectorShift)&clusterMask)

PUB getFilesize                  ' Return current file size
   return filesize

PUB getClusterSize               ' Return current cluster size in bytes
   return clusterByteSize

PRI flushMeta                    ' Flush metadata buffer if changed
   if dirty~ and lastRead <> -1
      if ldr.writeSDCard(lastRead,@metaBuf,sectorSize)
         abort -@writeSDErr      ' Update both FAT regions if necessary
      if lastRead => fat1 and lastRead < fat1 + sectorsPerFAT
         if ldr.writeSDCard(lastRead+sectorsPerFAT,@metaBuf,sectorSize)
            abort -@writeSDErr

PRI readMeta(n)                  ' Read metadata if necessary
   if n <> lastRead
      flushMeta
      if ldr.readSDCard(n,@metaBuf,sectorSize)
         abort -@readSDErr
      lastRead := n

PRI metaReadWord(n)              ' Return byte aligned word data
   bytemove(@result,@metaBuf+n,2)

PRI metaReadLong(n)              ' Return byte aligned long data
   bytemove(@result,@metaBuf+n,4)

PRI metaWriteWord(n,v)           ' Change byte aligned word data
   bytemove(@metaBuf+n,@v,2)
   dirty~~

PRI metaWriteLong(n,v)           ' Change byte aligned long data
   bytemove(@metaBuf+n,@v,4)
   dirty~~

PRI readCluster(clust)           ' Read current cluster, return offset
   fatPtr := (fat1 << sectorShift) + (clust << 1)
   readMeta(fatPtr >> sectorShift)
   return fatPtr & sectorMask

PRI nextCluster                  ' Return next cluster number
   result := metaReadWord(readCluster(fclust))
   writeLink := fatPtr           ' Update writeLink value
   if result < 2 or result => totalClusters
      abort -@badClusterErr

PRI fillDataBuffer               ' Fill data buffer with next block
   if floc => fileSize
      return -1                  ' Return -1 if no more blocks
   if frem == 0
      fclust := nextCluster      ' Go to next cluster if necessary
      frem := clusterByteSize <# filesize - floc
   if ldr.readSDCard(getDataBlock,@buffer,sectorSize)
      abort -@readSDErr
   result := sectorSize <# fileSize - floc
   floc += result                ' Return number of bytes read
   frem -= result                '  and adjust pointers as needed
   bufat := 0
   bufend := result

PRI freeClusters(clust)          ' Free entire cluster chain
   repeat while clust < $fff0
      if clust < 2               ' Check for valid cluster number
         abort -@badClusterErr
      result := readCluster(clust)
      clust := metaReadWord(result)
      metaWriteWord(result,0)    ' Get next cluster & set to zero
   flushMeta

PRI findNewCluster | clustStart, i, count
   clustStart := fclust & !sectorMask
   result := metaReadWord(readCluster(fclust))
   if result => 2 and result < totalClusters
      fclust := result
      frem := clusterByteSize
      writeLink := fatPtr
      return
   result := -1
   count := 2
   repeat
      readCluster(clustStart)
      repeat i from 0 to sectorSize-2 step 2
         if metaReadWord(i) == 0 ' Found free cluster
            result := clustStart + (i >> 1)
            if result < totalClusters ' Within volume?
               metaWriteWord(i,-1) ' Mark as last in file
               writeLink := fatPtr + i
               metaWriteWord(readCluster(writeLink),result)
               fclust := result ' Add new cluster to chain
               frem := clusterByteSize
               return
      clustStart += sectorSize >> 1
      if clustStart => totalClusters
         clustStart := 0         ' Wrap around to start of volume
         if count-- == 0
            return -@noSpaceErr

PUB flushBuffer(meta) | t        ' Flush current write buffer
   if readMode == 0
      abort -@notWriteErr        ' Must be write mode
   if result := bufat
      if frem < sectorSize       ' Find a new cluster
         findNewCluster
      if frem => sectorSize
         if ldr.writeSDCard(getDatablock,@buffer,sectorSize)
            abort -@writeSDErr
         if result == sectorSize ' Full buffer, clear it
            floc += result
            frem -= result
            bufat := 0
            bufend := result
   if result < 0 or meta         ' Update metadata even on error
      readMeta(dirEntry >> sectorShift)
      metaWriteLong((dirEntry&sectorMask)+28,floc+bufat)
      flushMeta
   if result < 0
      abort result

PRI lookupName(n1, n2, n3) | i, f1,f2,f3
   repeat i from 1 to 15         ' Look in each 32K EEPROM area
      result := def#bootAddr | (i << 15)
      if ldr.checkPresence(result)
         if ldr.readEEPROM(result+nameAddr,@f1,12)
            abort -@readEepromErr
         if n1 == f1 and n2 == f2 and n3 == f3
            return
   return -1

PRI directoryLookup(n,subDir) | free, i, p, attr
   free := 0                     ' Return address of found entry
   repeat dirEntry from rootDir to rootDirEnd - dirSize step dirSize
      readMeta(dirEntry >> sectorShift)
      p := dirEntry & sectorMask
      if metaBuf.byte[p] == 0    ' End of directory?
         if free == 0
            dirEntry := free     ' If deleted entry found, provide it
         return false            '  else provide end of directory
      if metaBuf.byte[p] == $e5 and free == 0
         free := result          ' Save first free entry
      repeat i from 0 to 10
         if byte[n+i] <> metaBuf.byte[p+i]
            quit
      attr := metaBuf.byte[p+$0b]
      if i < 11 or attr & $70 <> 0 ' Mismatch, volume label, or invalid
         next
      if attr & $80 == subDir    ' Find what we're looking for?
         fclust := metaReadWord(writeLink := p + $1a)
         fileSize := metaReadLong(p+$1c)
         return true
   dirEntry := free              ' Nothing found, return 1st free
   return false

PRI uc(c)                        ' Force lower case letters to upper
   if c => "a" and c =< "z"
      c -= 32
   return c

PRI dateTime                     ' Return current date and time
'   Get the current date and time, as a long, in the format required by
'   FAT16.  Currently, this returns a fixed date because there's no RTC.
   return constant(((2007-1980) << 25) + (1 << 21) + (26 << 16) + (12 << 11))