'' This is the support library for the Propeller Meta Compiler.
'' It's based on the Meta2 Compiler-Compiler written originally
'' for the IBM 360 by Michael Green and that, in turn, is based
'' on the Meta Compiler-Compiler presented at a 1963 ACM
'' conference on programming languages.

'' 2007-11-29 - Added basic symbol dictionary
'' 2007-11-30 - Added routines for binary output to EEPROM
'' 2007-12-02 - getDelimiter has option to check for delimiter
'' 2007-12-03 - backupSymbol backs up before last identifier
''              getContent also sets value
''              getLength returns symbol length
'' 2007-12-08 - Added outputInst to output instruction to EEPROM
''              Added pass through calls for read/write EEPROM
''              Added getKeyword to handle keyword tables
''              Added getOperator to handle operator tables
''              Added getSymBuf to return address of symbol
''              Added getEndfile to test for end of file
''              Added messageDec, messageHex, messageHex, messageOut
'' 2007-12-09 - Added outputSymbol to get symbol text from dictionary
'' 2007-12-14 - getSymbol corrected
'' 2007-12-15 - getContent corrected
'' 2007-12-16 - addToDict corrected, improved getSymbol, nextDictPos
''              Corrected output buffer addressing
'' 2007-12-17 - Corrected outputSymbol
'' 2007-12-18 - Modified outputSymbol to copy to symBuf as well

OBJ
   def  : "Definitions"          ' Common definitions
   mass : "MassStorage"          ' EEPROM and SD card interface
   key  : "Keyboard"             ' Keyboard I/O routines
   dsp  : "Display"              ' Display I/O routines
   vga  : "VgaDriver"            ' VGA driver

CON
   version   = 1
   release   = 001
   testlevel = 1

   maxSymbol = 128               ' Maximum symbol length
   maxBackup = 128               ' Maximum saved characters
   maxOutput = 128               ' EEPROM page size
   maxFileNm = 12                ' Filename length
   maxLine   = 80                ' Maximum line length for debugging

   bspChar   = $08               ' ASCII backspace
   tabChar   = $09               ' ASCII tab
   lfChar    = $0A               ' ASCII line feed
   crChar    = $0D               ' ASCII carriage return
   quote     = 34                ' ASCII double quote

VAR
   long inputAddr                ' Input EEPROM address
   long value                    ' Holds scanned numeric value
   long indent                   ' Number of spaces to indent
   long debug                    ' True if in debug mode
   long lineOriented             ' True if CR / LF is significant
   long commentType              ' "S" for Spin, "C" for C
   long ignoreCase               ' True if letter case ignored
   byte fifoInPtr
   byte fifoOutPtr               ' Input FIFO
   byte fifoBuffer[maxBackup]
   byte symPtr
   byte symBuf[maxSymbol+1]      ' Holds scanned symbol
   long dictBase, dictLast
   long dictAddr                 ' Dictionary EEPROM address
   byte dictPtr
   byte dictBuf[maxOutput]       ' Used for adding to dictionary
   long outputBase
   long outputAddr               ' Output EEPROM address
   byte outputPtr
   byte outputBuf[maxOutput]     ' Holds EEPROM output page
   byte debugInPtr
   byte debugInBuf[maxLine+1]    ' Holds current input line for debugging
   byte debugOutPtr
   byte debugOutBuf[maxLine+1]   ' Holds current output line for debugging

PUB start(s,d) | c, tp, cp       ' Initialize support library
   debug := d
   vga.start(def#vgaPins)        ' Start display driver
   dsp.str(string("Meta2 Library "))
   dsp.dec(version)
   dsp.out(".")
   if release < 100
      dsp.out("0")
   if release < 10
      dsp.out("0")
   dsp.dec(release)
   if testLevel > 0
      dsp.out("a"+testLevel-1)
   dsp.str(string(crChar,lfChar))
   key.start(def#keyPins)        ' Start keyboard driver
   if \mass.start
      dsp.str(string("Can't start I2C/SPI",crChar,lfChar))
      return false
   fifoInPtr := fifoOutPtr := 0
   debugInPtr := debugOutPtr := 0
   lineOriented := false
   commentType := "S"
   ignoreCase := true
   indent := 0
   if s <> 0                     ' File name supplied?
      if \mass.mountSDVol(mass#spiDO,mass#spiClk,mass#spiDI,mass#spiCS)
         dsp.str(string("Can't mount SD card",crChar,lfChar))
         return false
      if s == -1
         askFilename(string("Input"))
         if symBuf[0] <> 0       ' Use EEPROM if no file name
            ifnot copyToEEPROM(@symBuf)
               dsp.str(string("No such file",crChar,lfChar))
               return false
      else
         ifnot copyToEEPROM(s)   ' Use supplied file name
            dsp.str(string("Can't open '"))
            dsp.str(s)
            dsp.str(string("'",crChar,lfChar))
            return false
      mass.unmountSDVol
   inputAddr := mass#bootAddr + $10000 ' Source input at $10000
   outputBase := mass#bootAddr + $18000 ' Output at $18000
   outputAddr := outputBase
   outputPtr := 0
   dictBase := mass#bootAddr + $08000 ' Dictionary at $08000
   dictLast := dictBase
   bytefill(@dictBuf,$FF,maxOutput)
   dictBuf[0] := 0               ' Initialize the dictionary
   if \mass.writeEEPROM(dictBase,@dictBuf,maxOutput)
      abort string("Dictionary writeEEPROM error")
   if \mass.writeWait(dictBase)
      abort string("Dictionary writeWait error")
   dsp.str(string("Beginning of compilation",crChar,lfChar))
   return true

PRI copyToEEPROM(s) | c          ' Copy file to EEPROM
   if \mass.openFile(s,"r") <> 0
      return false               ' Can't open the file
   outputAddr := mass#bootAddr + $10000 ' Write to input
   repeat
      c := \mass.readFile(@outputBuf,maxOutput)
      if c == -1                 ' No more blocks to read
         quit
      if \mass.writeEEPROM(outputAddr,@outputBuf,c)
         abort string("copyToEEPROM - writeEEPROM error")
      if \mass.writeWait(outputAddr)
         abort string("copyToEEPROM - writeWait error")
      if c < maxOutput           ' Partial block read
         outputAddr += c
         quit
      outputAddr += maxOutput    ' On to next block
   c := 0                        ' Now write zero terminator
   if \mass.writeEEPROM(outputAddr,@c,1)
      abort string("copyToEEPROM - writeEEPROM EOF error")
   if \mass.writeWait(outputAddr)
      abort string("copyToEEPROM - writeWait EOF error")
   return true

PUB stop(s) | c                  ' Finish up processing
   dsp.str(string("End of compilation",crChar,lfChar))
   outputBuf[outputPtr++] := 0
   if \mass.writeEEPROM(outputAddr,@outputBuf,outputPtr)
      dsp.str(string("writeEEPROM error",crChar,lfChar))
      return false
   if \mass.writeWait(outputAddr)
      dsp.str(string("writeWait error",crChar,lfChar))
      return false
   if s <> 0                     ' File name supplied?
      if \mass.mountSDVol(mass#spiDO,mass#spiClk,mass#spiDI,mass#spiCS)
         dsp.str(string("Can't mount SD card",crChar,lfChar))
         return false
      if s == -1
         askFilename(string("Output"))
         if symBuf[0] <> 0       ' Use EEPROM if no file name
            ifnot copyFromEEPROM(@symBuf)
               return false
      else
         ifnot copyFromEEPROM(s)
            return false
      mass.unmountSDVol
   return true

PRI copyFromEEPROM(s)            ' Copy EEPROM to file
   if \mass.openFile(s,"w") <> 0
      dsp.str(string("Can't create file",crChar,lfChar))
      return false
   inputAddr := mass#bootAddr + $18000 ' Read from output
   outputBuf[maxOutput]~         ' Force zero byte terminator
   repeat
      if \mass.readEEPROM(inputAddr,@outputBuf,maxOutput)
         dsp.str(string("readEEPROM error",crChar,lfChar))
         return false
      if outputPtr := strsize(@outputBuf)
         inputAddr += outputPtr
         if \mass.writeFile(@outputBuf,outputPtr) < 0
            dsp.str(string("writeFile error",crChar,lfChar))
            return false
   until outputPtr < maxOutput
   \mass.closeFile               ' All done
   dsp.str(string("Output copied to file",crChar,lfChar))
   return true

PRI askFilename(s) | i           ' Get file name into symBuf
   dsp.str(s)
   dsp.str(string(" file? "))
   i := 0
   repeat
      case symBuf[i] := key.key
         bspChar:                ' Backspace
            if i
               dsp.str(string(bspChar," ",bspChar))
               i--
         crChar:                 ' Carriage return
            symBuf[i]~           ' Zero byte terminator
            dsp.str(string(crChar,lfChar))
            quit
         other:                  ' Other characters,
            if i < maxFileNm     '  save up to maxFilenm
               dsp.out(symBuf[i++])

PUB checkAbort(a) | c            ' Check for error exit
   if a == true
      return true                ' Successful completion
   if a == false
      a := string("Syntax error")
   dsp.str(a)
   dsp.str(string(crChar,lfChar))
   debugInBuf[debugInPtr]~       ' Zero terminator
   dsp.str(string(">>"))
   dsp.str(@debugInBuf)          ' Display debug buffer
   dsp.str(string("<<",crChar,lfChar))
   return false

PUB exit
   reboot

PUB messageOut(c)                ' Copy the character to the console
   dsp.out(c)

PUB message(s)                   ' Copy the string to the console
   dsp.str(s)

PUB messageDec(v)                ' Copy a decimal value to the console
   dsp.dec(v)

PUB messageHex(v,n)              ' Copy a hexadecimal value to the console
   dsp.hex(v,n)

PUB messageBin(v,n)              ' Copy a binary value to the console
   dsp.bin(v,n)

PUB messageNewline               ' Copy a newline to the console
   dsp.str(string(crChar,lfChar))

PUB setIgnoreCase(b)             ' Set value of ignoreCase
   ignoreCase := b

PRI newChar                      ' Return the next input character
   if fifoInPtr == fifoOutPtr
      if \mass.readEEPROM(inputAddr,@result,1)
         abort string("newChar - readEEPROM error")
      if result                  ' Zero byte marks end of input
         inputAddr++             ' If not end, advance pointer
      if (inputAddr & $3FFFF) > $17FFF
         abort string("newChar - end of EEPROM")
      if result > 0 and result <> lfChar
         if result == crChar     ' Save input line for debugging
            debugInBuf[debugInPtr]~
            debugInPtr := 0      ' Prepare for next input line
            if debug
               dsp.str(string(">>"))
               dsp.str(@debugInBuf)
               dsp.str(string("<<",crChar,lfChar))
         else
            if debugInPtr < maxLine
               debugInBuf[debugInPtr++] := result
      fifoBuffer[fifoInPtr++] := result
      fifoInPtr //= maxBackup    ' Save character in FIFO
      fifoOutPtr := fifoInPtr
   else
      result := fifoBuffer[fifoOutPtr++]
      fifoOutPtr //= maxBackup   ' Get next character out of FIFO

PUB getLength                    ' Get length of .ident or .content
   return symPtr                 '  or formatted decimal or hex value

PUB getSymBuf                    ' Get address of symbol buffer
   return @symBuf

PUB backupSymbol                 ' Back up to start of current symbol
   backupChar(symPtr)

PRI backupChar(n)                ' Back up n characters in input buffer
   fifoOutPtr := (fifoOutPtr + maxBackup - n) // maxBackup

PRI skipWhite | c, x, n          ' Skip white space
   repeat
      case c := newChar
         " ", tabChar:
            next
         lfChar, crChar:         ' If line oriented, CR/LF not whitespace
            if lineOriented
               backupChar(1)
               quit
         "/":                    ' Handle C type comments with /* ... */
            if commentType <> "C"
               backupChar(1)
               quit
            if newChar <> "*"
               backupChar(2)
               quit
            repeat
               case x := newChar
                  0: abort string("End of file in comment")
                  "*":
                     if newChar == "/"
                        quit
         "'":                    ' Handle quote to end of line comments
            if commentType <> "S"
               backupChar(1)
               quit
            repeat
               case x := newChar
                  0: abort string("End of file in comment")
                  lfChar, crChar:
                     backupChar(1)
                     quit
         "{":                    ' Handle nested braces comments
            if commentType <> "S"
               backupChar(1)
               quit
            n := 1
            repeat
               case x := newChar
                  0: abort string("End of file in comment")
                  "{":
                     n++         ' Count nesting level
                  "}":
                     ifnot --n   ' Still nested, continue skipping
                        quit
         other:
            backupChar(1)        ' Back up one character
            quit       

PUB getValue                     ' Return the numeric value just scanned in
   return value

PUB setValue(v)                  ' Set the value variable
   value := v

PUB getIndent                    ' Return column number of first non-whitespace
   skipWhite                     '  after a span of whitespace.  May not work
   return debugInPtr             '   if called when currently not in whitespace

PUB setLineOriented(b)           ' Set lineOriented
   lineOriented := b

PUB setCommentType(t)            ' Set commentType
   commentType := t

PUB getEndline | c               ' Match CR or LF or CR/LF if lineOriented
   skipWhite
   ifnot lineOriented
      return false               ' Never matches if not lineOriented
   case newChar
      crChar:
         if newChar <> lfChar
            backupChar(1)
         return true
      lfChar:
         return true
      other:
         backupChar(1)
         return false   

PUB getEndfile                   ' Match end of file, but leave
   skipWhite                     '  it positioned before it
   if newChar
      backupChar(1)
      return false
   backupChar(1)
   return true

PUB getLiteral(target) | i,s,c,d ' Match against supplied string
   skipWhite                     ' If address is negative, the string
   s := strsize(||target)        '  must be followed by a delimiter
   if s > maxBackup
      abort string("literal - maxBackup size")
   repeat i from 0 to s-1
      c := newChar
      d := byte[||target][i]
      if ignoreCase              ' Convert lower case to upper case
         if c => "a" and c =< "z"
            c &= %1101_1111
         if d => "a" and d =< "z"
            d &= %1101_1111
      if c <> d
         backupChar(i+1)
         return false
   if s < 0                      ' String must be followed by a delimiter
      case newChar
         "_", "0".."9", "A".."Z", "a".."z":
            backupChar(s+1)      ' If not, treat the match as unsuccessful
            return false
      backupChar(1)              ' If so, leave the delimiter there
   return true

PUB getIdent | c                 ' Does .id
   skipWhite
   case c := newChar
      "_", "A".."Z", "a".."z":   ' Initial letter or "_"
         if ignoreCase
            if c => "a" and c =< "z"
               c &= %1101_1111   ' Convert lower case to upper case
         symPtr := 0
         symBuf[symPtr++] := c   ' Save character
      other:
         backupChar(1)
         return false            ' Not identifier
   repeat
      case c := newChar
         "_", "0".."9", "A".."Z", "a".."z":
            if ignoreCase
               if c => "a" and c =< "z"
                  c &= %11011111 ' Convert lower case to upper case
            if symPtr < maxSymbol ' Save 1st part of identifier
               symBuf[symPtr++] := c
            next                 ' Look for more
         other:
            symBuf[symPtr]~      ' Zero byte terminator
            backupChar(1)
            return true

PRI checkNumber(base,digit)      ' Do value := value * base + digit
   if value ** base <> 0         ' Check for pending 32 bit overflow
      abort string("checkNumber - overflow base")
   value *= base
   if (value + digit) & $8000_0000 == 0 and value & $8000_0000 <> 0
      abort string("checkNumber - overflow digit")
   return value := value + digit

PUB getNumber | c, digits        ' Does .number
   skipWhite
   value := 0
   digits := 0
   case c := newChar
      "%":                       ' Binary or quaternary value
         if (c := newChar) == "%"
            repeat               ' Quaternary value
               case c := newChar
                  "_":
                     next
                  "0".."3":
                     checkNumber(4,c-"0")
                     digits++
                     next
                  other:
                     backupChar(1)
                     return digits > 0
         else
            fifoOutPtr := (fifoInPtr - 1) // maxBackup
            repeat               ' Binary value
               case c := newChar
                  "_":
                     next
                  "0".."1":
                     checkNumber(2,c-"0")
                     digits++
                     next
                  other:
                     backupChar(1)
                     return digits > 0
      "$":                       ' Hexadecimal value
         repeat
            case c := newChar
               "_":
                  next
               "0".."9":
                  checkNumber(16,c-"0")
                  digits++
                  next
               "A".."F":
                  checkNumber(16,c-"A"+10)
                  digits++
                  next
               "a".."f":
                  checkNumber(16,c-"a"+10)
                  digits++
                  next
               other:
                  backupChar(1)
                  return digits > 0
      "0".."9":
         checkNumber(10,c-"0")
         digits++
      other:
         backupChar(1)
         return digits > 0
   repeat
      case c := newChar
         "_":
            next
         "0".."9":
            checkNumber(10,c-"0")
            digits++
            next
         other:
            backupChar(1)
            return digits > 0

PRI storeString(c,p)
   symBuf[p++] := c
   if p == maxOutput
      abort string("string - too long")

PUB getString | p, c, s          ' Does .string
   skipWhite
   if newChar <> quote           ' Open quote?
      backupChar(1)
      return false
   p := 0
   s := 1                        ' State = Just started
   repeat
      case c := newChar
         lfChar, crChar:         ' End of line in string
            abort string("string - end of line")
         quote:
            if newChar == quote  ' Escaped quote
               if s == 2         ' End a substring with a quote
                  storeString(quote,p++)
               if s <> 1         ' Not at the start of the string
                  storeString(",",p++)
               storeString("3",p++)
               storeString("4",p++)
               s := 3            ' Quote placed in string
            else
               backupChar(1)
               if s == 2         ' End a substring with a quote
                 storeString(quote,p++)
               symBuf[p]~        ' Zero byte terminator
               return true
         other:
            if c < " "           ' Invalid character
               abort string("string - invalid")
            if s == 3            ' Previous character was a quote
               storeString(",",p++)
            if s <> 2            ' Start new substring with quote
               storeString(quote,p++)
            storeString(c,p++)
            s := 2               ' At least one non-quote stored

PUB getContent | c               ' Like .string, but string content only
   skipWhite
   if newChar <> quote           ' Open quote?
      backupChar(1)
      return false
   symPtr := 0
   repeat
      case c := newChar
         lfChar, crChar:         ' End of line in string
            abort string("string - end of line")
         quote:
            if newChar == quote ' Escaped quote
               storeString(c,symPtr++)
            else
               backupChar(1)
               symBuf[symPtr]~   ' Zero byte terminator
               if symPtr =< 4
                  value := 0     ' Get numeric value of string
                  if symPtr > 0
                     bytemove(@value,@symBuf,symPtr)
               return true
         other:
            if c < " "           ' Invalid character
               abort string("string - invalid")
            storeString(c,symPtr++)

PUB convertHex(digits) | v       ' Convert value to hexadecimal string
  symPtr := 0
  symBuf[symPtr++] := "$"        ' Leading "$"
  v := value << (8 - digits) << 2
  repeat digits
    symBuf[symPtr++] := lookupz((v <-= 4) & $F : "0".."9", "A".."F")
  symBuf[symPtr]~                ' Zero byte terminator

PUB convertDec | i, x, v         ' Convert value to decimal string
   symPtr := x := 0
   if (v := value) < 0           ' Leading sign if negative
      -v
      symBuf[symPtr++] := "-"
   i := 1_000_000_000
   repeat 10                     ' Up to 10 digits
      if v => i
         symBuf[symPtr++] := v / i + "0"
         v //= i
         x~~                     ' Always have at least 1 digit
      elseif x or i == 1
         symBuf[symPtr++] := "0"
      i /= 10
   symBuf[symPtr]~               ' Zero byte terminator

DAT
spaces    byte  "                "  ' A total of 32 spaces
          byte  "                "
          byte  0

PUB outputIndent                 ' Indent # of spaces requested
   output(@spaces+(32-indent))

PUB adjustIndent(adj)            ' Adjust # of spaces to indent
   indent += adj
   if indent < 0 or indent > 32
      abort string("adjustIndent - out of range")

PUB outputNewline                ' Output a carriage return / line feed
   if debug
      debugOutBuf[debugOutPtr]~  ' Zero termination byte
      dsp.str(string("|"))
      dsp.str(@debugOutBuf)      ' Show saved output line
      dsp.str(string("|",crChar,lfChar))
   output(string(crChar,lfChar))
   debugOutPtr := 0

PUB output(s) | c                ' Copy string to output
   if s == 0
      s := @symBuf               ' Use symbol buffer if null
   repeat while c := byte[s++]
      if debugOutPtr < maxLine   ' Save output line for debugging
         debugOutBuf[debugOutPtr++] := c
      outputByte(c)

PUB outputByte(v)                ' Copy byte value to output
   outputBuf[outputPtr++] := v
   if outputPtr == maxOutput
      if \mass.writeEEPROM(outputAddr,@outputBuf,maxOutput)
         abort string("outputByte - writeEEPROM error")
      if \mass.writeWait(outputAddr)
         abort string("outputByte - writeWait error")
      outputAddr += maxOutput
      if (outputAddr & $7FFF) == 0
         abort string("outputByte - end of EEPROM")
      outputPtr := 0
     
PUB outputWord(v)                ' Copy word value to output
   outputByte(v & $FF)           '  LSB first, then MSB
   outputByte(v >> 8)

PUB outputLong(v)                ' Copy long value to output
   outputWord(v & $FFFF)         '  LSB first, MSB last
   outputWord(v >> 16)

PUB outputInst(opcode, flags, cond, dest, srce) ' Output instruction
   outputLong(opcode << 26 | flags << 22 | cond << 18 | dest << 9 | srce)

PUB getOutputAddr                ' Return current output EEPROM address
   return outputAddr + outputPtr - outputBase

PUB readLong(addr)               ' Read a long from EEPROM or the buffer
   addr += outputBase
   if addr => outputAddr + outputPtr
      abort string("readLong - address beyond EOF")
   if addr < outputAddr
      if \mass.readEEPROM(addr,@result,4)
         abort string("readLong - readEEPROM error")
   else
      bytemove(@result,@outputBuf+(addr-outputAddr),4)

PUB writeLong(addr, val)         ' Write a long to EEPROM or the buffer
   addr += outputBase
   if addr => outputAddr + outputPtr
      abort string("writeLong - address beyond EOF")
   if addr < outputAddr
      if \mass.writeEEPROM(addr,@val,4)
         abort string("writeLong - writeEEPROM error")
      if \mass.writeWait(addr)
         abort string("writeLong - writeWait error")
   else
      bytemove(@outputBuf+(addr-outputAddr),@val,4)

PUB str(s)                       ' Display supplied string
   dsp.str(s)

PUB dec(v)                       ' Display signed decimal number
   dsp.dec(v)

PUB hex(v,d)                     ' Display hexadecimal number of given width
   dsp.hex(v,d)

PUB bin(v,d)                     ' Display binary number of given width
   dsp.bin(v,d)

PUB out(c)                       ' Display given character
   dsp.out(c)

PRI addToDict(c)                 ' Add character to dictionary
   if dictPtr == maxOutput
      if \mass.writeEEPROM(dictAddr,@dictBuf,maxOutput)
         abort string("addToDict - writeEEPROM error")
      if \mass.writeWait(dictAddr)
         abort string("addToDict - writeWait error")
      dictLast := dictAddr
      dictAddr += maxOutput
      dictPtr := 0               ' Check for dictionary full
      if dictAddr-dictBase > $07FFF
         abort string("addToDict - end of EEPROM")
      bytefill(@dictBuf,$FF,maxOutput)
   dictBuf[dictPtr++] := c

PRI nextDictPos                  ' Move to next dictionary position
   if ++dictPtr == maxOutput
      dictAddr += maxOutput
      dictPtr := 0
      if dictAddr > dictLast
         abort string("nextDictPos - end of dictionary")
      if \mass.readEEPROM(dictAddr,@dictBuf,maxOutput)
         abort string("nextDictPos - readEEPROM error")

PUB getKeyword(d,s) | len, sym   ' Lookup symbol in supplied table
   if s == 0
      s := @symBuf
   sym := strsize(s)
   repeat while byte[d]
      len := strsize(d)
      if sym <> len              ' Lengths don't match
         d += len + 1
         result++
         next
      ifnot strcomp(d,s)         ' Symbols not equal
         d += len + 1
         result++
         next
      return                     ' Return symbol number
   return -1                     '  or -1 if not found
   
PUB getSymbol(s) | c, n, p, x, z ' Lookup symbol in dictionary
   if s == 0
      s := @symBuf               ' Use symbol buffer if null
   dictAddr := dictBase
   dictPtr := 0
   if \mass.readEEPROM(dictAddr,@dictBuf,maxOutput)
      abort string("getSymbol - readEEPROM error")
   c := strsize(s)               ' Get length of symbol
   repeat while x := dictBuf[dictPtr]
      nextDictPos                ' Get length of dictionary entry
      z := c == x                '  and compare to length of symbol
      p := 0
      repeat x                   ' Compare symbol to dictionary entry
         z := z and dictBuf[dictPtr] == byte[s][p++]
         nextDictPos
      if z                       ' Return dictionary entry number
         return
      result++
   addToDict(c)                  ' Store length of symbol
   repeat c                      '  followed by the symbol
      addToDict(byte[s++])
   addToDict(0)                  ' Zero marks end of dictionary
   if \mass.writeEEPROM(dictAddr,@dictBuf,maxOutput)
      abort string("getSymbol - writeEEPROM error")
   if \mass.writeWait(dictAddr)
      abort string("getSymbol - writeWait error")
   dictLast := dictAddr
   return                        ' Return dictionary entry number

PUB outputSymbol(n,copy) | d, c  ' Output symbol from dictionary
   c := 0                        '  as zero terminated string
   d := dictBase
   repeat
      if \mass.readEEPROM(d++,@result,1)
         abort string("outputSymbol - readEEPROM error")
      if result == 0
         return                  ' Return 0 if not found
      if n-- == 0
         quit
      d += result
   symPtr := 0                   ' Copy symbol to symBuf and
   repeat result                 '  optionally copy into output
      if \mass.readEEPROM(d++,@c,1)
         abort string("outputSymbol - readEEPROM error")
      symBuf[symPtr++] := c
      if copy
         outputByte(c)
   symBuf[symPtr] := 0           ' Add zero terminator
   if copy
      outputByte(0)              ' Return the length of the symbol
      
PUB getOperator(d) | c, i, s     ' Lookup operator in supplied table that
   skipWhite                     '  must be in order by length with the
   repeat while byte[d]          '   longest first.  Return the operator
      s := strsize(d)            '    number or -1 if not found
      repeat i from 0 to s-1
         c := newChar
         if c <> byte[d][i]      ' Mismatch
            backupChar(i+1)
            quit
         if i == s-1             ' Match, return operator number
            return
      d += s + 1                 ' Try next in list
      result++
   return -1                     ' No match